@@ -26,8 +26,8 @@
 #include <stdio.h>
 #include <fcntl.h>
 #include <string.h>
-#include <unistd.h>
 #include <syslog.h>
+#include <unistd.h>
 #include <signal.h>
 #include <getopt.h>
 #include <locale.h>
@@ -170,7 +170,7 @@ main (int argc, char *argv[])
 	    {
 	      char **gp;
 
-	      for (gp = grp->gr_mem; *gp != NULL; *gp++)
+	      for (gp = grp->gr_mem; *gp != NULL; gp++)
 		{
 		  if (strcmp (pw->pw_name, *gp) == 0)
 		    {
