@@ -42,6 +42,7 @@
 #define E_BAD_ENTRY 2
 #define E_NO_FILE 3
 #define E_PWDBUSY 4
+#define E_FAILURE 5
 
 #ifndef _
 #define _(String) gettext (String)
@@ -70,7 +71,7 @@ print_help (const char *program)
   fputs (_("  -q, --quiet    Don't print warnings, only errors\n"), stdout);
   fputs (_("  -r, --read-only Run in read-only mode, don't make changes\n"),
 	 stdout);
-  fputs (_("  -s, --sort     Sort the password file, not checks done\n"),
+  fputs (_("  -s, --sort     Sort the password file, no checks are done\n"),
 	 stdout);
   fputs (_("      --help     Give this help list\n"), stdout);
   fputs (_("  -u, --usage    Give a short usage message\n"), stdout);
@@ -688,6 +689,232 @@ loop_over_shadow_file (int quiet)
   return result;
 }
 
+/* XXX move n library.  */
+static struct passwd *
+files_getpwent (void)
+{
+  enum nss_status status;
+  int buflen = 256;
+  char *buffer = NULL;
+  struct passwd *resultbuf = malloc (sizeof (struct passwd));
+
+  if (buffer == NULL)
+    buffer = malloc (buflen);
+
+  while ((status = files_getpwent_r (resultbuf, buffer, buflen, &errno))
+         == NSS_STATUS_TRYAGAIN && errno == ERANGE)
+    {
+      errno = 0;
+      buflen += 256;
+      buffer = realloc (buffer, buflen);
+    }
+  if (status == NSS_STATUS_SUCCESS)
+    return resultbuf;
+  else
+    return NULL;
+}
+
+struct passwd_list {
+  struct passwd *pwd;
+  struct passwd_list *next;
+};
+
+static int
+sort_passwd_file (void)
+{
+  struct passwd *pwd;
+  struct passwd_list *ptr = NULL;
+  int retval = 0;
+
+  while ((pwd = files_getpwent ()))
+    {
+      if (pwd->pw_name[0] == '+' || pwd->pw_name[0] == '-')
+	break;
+
+      if (ptr == NULL)
+	{
+	  ptr = malloc (sizeof (struct passwd_list));
+	  if (ptr == NULL)
+	    {
+	      fputs ("running out of memory!\n", stderr);
+	      exit (E_FAILURE);
+	    }
+
+	  ptr->pwd = pwd;
+	  ptr->next = NULL;
+	}
+      else
+	{
+	  struct passwd_list *next = ptr;
+
+	  while (next->next && next->pwd->pw_uid < pwd->pw_uid)
+	    next = next->next;
+
+	  if (next->next == NULL)
+	    {
+	      next->next = malloc (sizeof (struct passwd_list));
+
+	      if (next->next == NULL)
+		{
+		  fputs ("running out of memory!\n", stderr);
+		  exit (E_FAILURE);
+		}
+	      next->next->next = NULL;
+
+	      if (pwd->pw_uid > next->pwd->pw_uid)
+		next->next->pwd = pwd;
+	      else
+		{
+		  next->next->pwd = next->pwd;
+		  next->pwd = pwd;
+		}
+	    }
+	  else
+	    {
+	      struct passwd_list *tmp;
+	      tmp = malloc (sizeof (struct passwd_list));
+
+	      if (tmp == NULL)
+		{
+		  fputs ("running out of memory!\n", stderr);
+		  exit (E_FAILURE);
+		}
+	      tmp->next = next->next;
+	      next->next = tmp;
+	      tmp->pwd = next->pwd;
+	      next->pwd = pwd;
+	    }
+	}
+    }
+
+  const char *file_tmp = "/passwd.tmpXXXXXX";
+  char *passwd_tmp = alloca (strlen (files_etc_dir) + strlen (file_tmp) + 1);
+  char *passwd_orig = alloca (strlen (files_etc_dir) + 8);
+  char *passwd_old = alloca (strlen (files_etc_dir) + 12);
+  struct stat passwd_stat;
+  FILE *oldgf, *newgf;
+  int newgf_fd;
+  char *cp;
+
+
+  cp = stpcpy (passwd_tmp, files_etc_dir);
+  strcpy (cp, file_tmp);
+  cp = stpcpy (passwd_orig, files_etc_dir);
+  strcpy (cp, "/passwd");
+  cp = stpcpy (passwd_old, passwd_orig);
+  strcpy (cp, ".old");
+
+  if ((oldgf = fopen (passwd_orig, "r")) == NULL)
+    {
+      fprintf (stderr, _("Can't open %s: %m\n"), passwd_orig);
+      retval = -1;
+      goto error_passwd;
+    }
+  if (fstat (fileno (oldgf), &passwd_stat) < 0)
+    {
+      fprintf (stderr, _("Can't stat %s: %m\n"), passwd_orig);
+      fclose (oldgf);
+      retval = -1;
+      goto error_passwd;
+    }
+
+#ifdef WITH_SELINUX
+  security_context_t prev_context;
+  if (set_default_context (passwd_orig, &prev_context) < 0)
+    {
+      fclose (oldgf);
+      retval = -1;
+      goto error_passwd;
+    }
+#endif
+  /* Open a temp passwd file */
+  newgf_fd = mkstemp (passwd_tmp);
+#ifdef WITH_SELINUX
+  restore_default_context (prev_context);
+#endif
+  if (newgf_fd == -1)
+    {
+      fprintf (stderr, _("Can't create %s temp file: %m\n"),
+	       passwd_orig);
+      fclose (oldgf);
+      retval = -1;
+      goto error_passwd;
+    }
+  fchmod (newgf_fd, passwd_stat.st_mode);
+  fchown (newgf_fd, passwd_stat.st_uid, passwd_stat.st_gid);
+  if (copy_xattr (passwd_orig, passwd_tmp) != 0)
+    {
+      fclose (oldgf);
+      close (newgf_fd);
+      unlink (passwd_tmp);
+      retval = -1;
+      goto error_passwd;
+    }
+
+  newgf = fdopen (newgf_fd, "w+");
+  if (newgf == NULL)
+    {
+      fprintf (stderr, _("Can't open %s: %m\n"), passwd_tmp);
+      fclose (oldgf);
+      close (newgf_fd);
+      unlink (passwd_tmp);
+      retval = -1;
+      goto error_passwd;
+    }
+
+  while (ptr != NULL)
+    {
+      /* write the passwd entry to tmp file */
+      if (putpwent (ptr->pwd, newgf) < 0)
+	goto write_error_passwd;
+      ptr = ptr->next;
+    }
+
+  /* Check if we have entries starting with +/- and copy
+     the rest of the passwd file without sorting it. */
+  if (pwd != NULL)
+    {
+      /* write the passwd entry to tmp file */
+      if (putpwent (pwd, newgf) < 0)
+	goto write_error_passwd;
+      while ((pwd = files_getpwent ()))
+      /* write the passwd entry to tmp file */
+      if (putpwent (pwd, newgf) < 0)
+	{
+	write_error_passwd:
+	  fprintf (stderr,
+		   _("Error while writing temporary passwd file: %m\n"));
+	  fclose (oldgf);
+	  fclose (newgf);
+	  retval = -1;
+	  goto error_passwd;
+	}
+    }
+
+  if (fclose (oldgf) != 0)
+    {
+      fprintf (stderr,
+	       _("Error while closing old passwd file: %m\n"));
+      fclose (newgf);
+      retval = -1;
+      goto error_passwd;
+    }
+  if (fclose (newgf) != 0)
+    {
+      fprintf (stderr,
+	       _("Error while closing temporary passwd file: %m\n"));
+      retval = -1;
+      goto error_passwd;
+    }
+  unlink (passwd_old);
+  link (passwd_orig, passwd_old);
+  rename (passwd_tmp, passwd_orig);
+ error_passwd:
+  unlink (passwd_tmp);
+
+  return retval;
+}
+
 int
 main (int argc, char *argv[])
 {
@@ -773,5 +1000,8 @@ main (int argc, char *argv[])
       return E_PWDBUSY;
     }
 
-  return loop_over_passwd_file (quiet) + loop_over_shadow_file (quiet);
+  if (sort)
+    return sort_passwd_file ();
+  else
+    return loop_over_passwd_file (quiet) + loop_over_shadow_file (quiet);
 }
